from code42cli.args import ArgConfig
from code42cli.commands import Command, SubcommandLoader
from code42cli.cmds.alerts.extraction import extract
from code42cli.cmds.search_shared import args, logger_factory
from code42cli.cmds.search_shared.enums import (
    AlertFilterArguments,
    AlertState,
    AlertSeverity,
    ServerProtocol,
    RuleType,
)
from code42cli.cmds.search_shared.cursor_store import AlertCursorStore


class MainAlertsSubcommandLoader(SubcommandLoader):
    PRINT = u"print"
    WRITE_TO = u"write-to"
    SEND_TO = u"send-to"
    CLEAR_CHECKPOINT = u"clear-checkpoint"

    def load_commands(self):
        """Sets up the `alerts` subcommand with all of its subcommands."""
        usage_prefix = u"code42 alerts"

        print_func = Command(
            self.PRINT,
            u"Print alerts to stdout",
            u"{} {}".format(usage_prefix, u"print <optional-args>"),
            handler=print_out,
            arg_customizer=_load_search_args,
            use_single_arg_obj=True,
        )

        write = Command(
            self.WRITE_TO,
            u"Write alerts to the file with the given name.",
            u"{} {}".format(usage_prefix, u"write-to <filename> <optional-args>"),
            handler=write_to,
            arg_customizer=_load_write_to_args,
            use_single_arg_obj=True,
        )

        send = Command(
            self.SEND_TO,
            u"Send alerts to the given server address.",
            u"{} {}".format(usage_prefix, u"send-to <server-address> <optional-args>"),
            handler=send_to,
            arg_customizer=_load_send_to_args,
            use_single_arg_obj=True,
        )

        clear = Command(
            self.CLEAR_CHECKPOINT,
            u"Remove the saved alert checkpoint from 'incremental' (-i) mode.",
            u"{} {}".format(usage_prefix, u"clear-checkpoint <optional-args>"),
            handler=clear_checkpoint,
        )

        return [print_func, write, send, clear]


def clear_checkpoint(sdk, profile):
    """Removes the stored checkpoint that keeps track of the last alert retrieved for the given profile..
        To use, run `code42 alerts clear-checkpoint`.
        This affects `incremental` mode by causing it to behave like it has never been run before.
    """
    AlertCursorStore(profile.name).replace_stored_cursor_timestamp(None)


def print_out(sdk, profile, args):
    """Activates 'print' command. It gets alerts and prints them to stdout."""
    logger = logger_factory.get_logger_for_stdout(args.format)
    extract(sdk, profile, logger, args)


def write_to(sdk, profile, args):
    """Activates 'write-to' command. It gets alerts and writes them to the given file."""
    logger = logger_factory.get_logger_for_file(args.output_file, args.format)
    extract(sdk, profile, logger, args)


def send_to(sdk, profile, args):
    """Activates 'send-to' command. It getsalerts and logs them to the given server."""
    logger = logger_factory.get_logger_for_server(args.server, args.protocol, args.format)
    extract(sdk, profile, logger, args)


def _load_write_to_args(arg_collection):
    output_file = ArgConfig(u"output_file", help=u"The name of the local file to send output to.")
    arg_collection.append(u"output_file", output_file)
    _load_search_args(arg_collection)


def _load_send_to_args(arg_collection):
    send_to_args = {
        u"server": ArgConfig(u"server", help=u"The server address to send output to."),
        u"protocol": ArgConfig(
            u"-p",
            u"--protocol",
            choices=ServerProtocol(),
            default=ServerProtocol.UDP,
            help=u"Protocol used to send logs to server.",
        ),
    }

    arg_collection.extend(send_to_args)
    _load_search_args(arg_collection)


def _load_search_args(arg_collection):
    filter_args = {
        AlertFilterArguments.SEVERITY: ArgConfig(
            u"--{}".format(AlertFilterArguments.SEVERITY),
            nargs=u"+",
            help=u"Filter alerts by severity. Defaults to returning all severities. Available choices={0}.".format(
                list(AlertSeverity())
            ),
        ),
        AlertFilterArguments.STATE: ArgConfig(
            u"--{}".format(AlertFilterArguments.STATE),
            help=u"Filter alerts by state. Defaults to returning all states. Available choices={0}.".format(
                list(AlertState())
            ),
        ),
        AlertFilterArguments.ACTOR: ArgConfig(
            u"--{}".format(AlertFilterArguments.ACTOR.replace("_", "-")),
            metavar=u"ACTOR",
            help=u"Filter alerts by including the given actor(s) who triggered the alert. Args must match actor username exactly.",
            nargs=u"+",
        ),
        AlertFilterArguments.ACTOR_CONTAINS: ArgConfig(
            u"--{}".format(AlertFilterArguments.ACTOR_CONTAINS.replace("_", "-")),
            metavar=u"ACTOR",
            help=u"Filter alerts by including actor(s) whose username contains the given string.",
            nargs=u"+",
        ),
        AlertFilterArguments.EXCLUDE_ACTOR: ArgConfig(
            u"--{}".format(AlertFilterArguments.EXCLUDE_ACTOR.replace("_", "-")),
            metavar=u"ACTOR",
            help=u"Filter alerts by excluding the given actor(s) who triggered the alert. Args must match actor username exactly.",
            nargs=u"+",
        ),
        AlertFilterArguments.EXCLUDE_ACTOR_CONTAINS: ArgConfig(
            u"--{}".format(AlertFilterArguments.EXCLUDE_ACTOR_CONTAINS.replace("_", "-")),
            metavar=u"ACTOR",
            help=u"Filter alerts by excluding actor(s) whose username contains the given string.",
            nargs=u"+",
        ),
        AlertFilterArguments.RULE_NAME: ArgConfig(
            u"--{}".format(AlertFilterArguments.RULE_NAME.replace("_", "-")),
            metavar=u"RULE_NAME",
            help=u"Filter alerts by including the given rule name(s).",
            nargs=u"+",
        ),
        AlertFilterArguments.EXCLUDE_RULE_NAME: ArgConfig(
            u"--{}".format(AlertFilterArguments.EXCLUDE_RULE_NAME.replace("_", "-")),
            metavar=u"RULE_NAME",
            help=u"Filter alerts by excluding the given rule name(s).",
            nargs=u"+",
        ),
        AlertFilterArguments.RULE_ID: ArgConfig(
            u"--{}".format(AlertFilterArguments.RULE_ID.replace("_", "-")),
            metavar=u"RULE_ID",
            help=u"Filter alerts by including the given rule id(s).",
            nargs=u"+",
        ),
        AlertFilterArguments.EXCLUDE_RULE_ID: ArgConfig(
            u"--{}".format(AlertFilterArguments.EXCLUDE_RULE_ID.replace("_", "-")),
            metavar=u"RULE_ID",
            help=u"Filter alerts by excluding the given rule id(s).",
            nargs=u"+",
        ),
        AlertFilterArguments.RULE_TYPE: ArgConfig(
            u"--{}".format(AlertFilterArguments.RULE_TYPE.replace("_", "-")),
            metavar=u"RULE_TYPE",
            help=u"Filter alerts by including the given rule type(s). Available choices={0}.".format(
                list(RuleType())
            ),
            nargs=u"+",
        ),
        AlertFilterArguments.EXCLUDE_RULE_TYPE: ArgConfig(
            u"--{}".format(AlertFilterArguments.EXCLUDE_RULE_TYPE.replace("_", "-")),
            metavar=u"RULE_TYPE",
            help=u"Filter alerts by excluding the given rule type(s). Available choices={0}.".format(
                list(RuleType())
            ),
            nargs=u"+",
        ),
        AlertFilterArguments.DESCRIPTION: ArgConfig(
            u"--{}".format(AlertFilterArguments.DESCRIPTION),
            help=u"Filter alerts by description. Does fuzzy search by default.",
        ),
    }
    search_args = args.create_search_args(search_for=u"alerts", filter_args=filter_args)
    arg_collection.extend(search_args)



import os

from code42cli.commands import Command
from code42cli import MAIN_COMMAND
from code42cli.commands import Command, SubcommandLoader
from code42cli.bulk import generate_template, BulkCommandType
from code42cli.cmds.alerts.rules.user_rule import (
    add_user,
    remove_user,
    get_rules,
    add_bulk_users,
    remove_bulk_users,
    show_rule,
)


def _customize_add_arguments(argument_collection):
    rule_id = argument_collection.arg_configs[u"rule_id"]
    rule_id.set_help(u"Observer ID of the rule to be updated.")
    username = argument_collection.arg_configs[u"username"]
    username.add_short_option_name("-u")
    username.set_help(u"The username of the user to add to the alert rule.")


def _customize_remove_arguments(argument_collection):
    rule_id = argument_collection.arg_configs[u"rule_id"]
    rule_id.set_help(u"Observer ID of the rule to be updated.")
    username = argument_collection.arg_configs[u"username"]
    username.add_short_option_name("-u")
    username.set_help(u"The username of the user to remove from the alert rule.")


def _customize_list_arguments(argument_collection):
    rule_id = argument_collection.arg_configs[u"rule_id"]
    rule_id.set_help(u"Observer ID of the rule.")


def _customize_bulk_add_arguments(argument_collection):
    _customize_bulk_arguments(argument_collection, u"adding")


def _customize_bulk_remove_arguments(argument_collection):
    _customize_bulk_arguments(argument_collection, u"removing")


def _customize_bulk_arguments(argument_collection, action):
    file_name = argument_collection.arg_configs[u"file_name"]
    file_name.set_help(
        u"The path to the csv file with columns 'rule_id,username' "
        u"for bulk {} users to the alert rule.".format(action)
    )


def _generate_template_file(cmd, path=None):
    """Generates a template file a user would need to fill-in for bulk operating.

    Args:
        cmd (str or unicode): An option from the `BulkCommandType` enum specifying which type of 
            file to generate.
        path (str or unicode, optional): A path to put the file after it's generated. If None, will 
            use the current working directory. Defaults to None.
    """
    handler = None
    filename = u"alert_rule.csv"
    if cmd == BulkCommandType.ADD:
        handler = add_user
        filename = u"add_users_to_{}".format(filename)
    elif cmd == BulkCommandType.REMOVE:
        handler = remove_user
        filename = u"remove_users_from_{}".format(filename)
    if not path:
        path = os.path.join(os.getcwd(), filename)
    generate_template(handler, path)


def _load_bulk_generate_template_description(argument_collection):
    cmd_type = argument_collection.arg_configs[u"cmd"]
    cmd_type.set_help(u"The type of command the template will be used for.")
    cmd_type.set_choices(BulkCommandType())


class AlertRulesBulkSubcommandLoader(SubcommandLoader):
    GENERATE_TEMPLATE = u"generate-template"
    ADD = u"add"
    REMOVE = u"remove"

    def load_commands(self):
        usage_prefix = u"{} alert-rules bulk".format(MAIN_COMMAND)

        generate_template_cmd = Command(
            self.GENERATE_TEMPLATE,
            u"Generate the necessary csv template for bulk actions.",
            u"{} generate-template <cmd> <optional args>".format(usage_prefix),
            handler=_generate_template_file,
            arg_customizer=_load_bulk_generate_template_description,
        )

        bulk_add = Command(
            self.ADD,
            u"Add users to alert rules. " u"CSV file format: `rule_id,username`.",
            u"{} add <filename>".format(usage_prefix),
            handler=add_bulk_users,
            arg_customizer=_customize_bulk_add_arguments,
        )

        bulk_remove = Command(
            self.REMOVE,
            u"Remove users from alert rules. " u"CSV file format: `rule_id,username`.",
            u"{} remove <filename>".format(usage_prefix),
            handler=remove_bulk_users,
            arg_customizer=_customize_bulk_remove_arguments,
        )

        return [generate_template_cmd, bulk_add, bulk_remove]


class AlertRulesSubcommandLoader(SubcommandLoader):
    ADD_USER = u"add-user"
    REMOVE_USER = u"remove-user"
    LIST = u"list"
    SHOW = u"show"
    BULK = u"bulk"

    def __init__(self, root_command_name):
        super(AlertRulesSubcommandLoader, self).__init__(root_command_name)
        self._bulk_subcommand_loader = AlertRulesBulkSubcommandLoader(self.BULK)

    def load_commands(self):
        usage_prefix = u"code42 alert-rules"

        add = Command(
            self.ADD_USER,
            u"Add a user to an alert rule.",
            u"{} add-user --rule-id <id>  --username <username>".format(usage_prefix),
            handler=add_user,
            arg_customizer=_customize_add_arguments,
        )

        remove = Command(
            self.REMOVE_USER,
            u"Remove a user from an alert rule.",
            u"{} remove-user --rule-id <rule-id> --username <username>".format(usage_prefix),
            handler=remove_user,
            arg_customizer=_customize_remove_arguments,
        )

        list_rules = Command(
            self.LIST,
            u"Fetch existing alert rules.",
            u"{} list".format(usage_prefix),
            handler=get_rules,
        )

        show = Command(
            self.SHOW,
            u"Print out detailed alert rule criteria.",
            u"{} show <rule-id>".format(usage_prefix),
            handler=show_rule,
            arg_customizer=_customize_list_arguments,
        )

        bulk = Command(
            self.BULK,
            u"Tools for executing bulk commands.",
            subcommand_loader=self._bulk_subcommand_loader,
        )

        return [add, remove, list_rules, show, bulk]



import logging

from c42eventextractor.logging.formatters import (
    FileEventDictToCEFFormatter,
    FileEventDictToJSONFormatter,
    FileEventDictToRawJSONFormatter,
)
from c42eventextractor.logging.handlers import NoPrioritySysLogHandlerWrapper

from code42cli.cmds.search_shared.enums import OutputFormat
from code42cli.util import get_url_parts
from code42cli.logger import (
    logger_has_handlers,
    logger_deps_lock,
    add_handler_to_logger,
    get_main_cli_logger,
    get_logger_for_stdout as get_stdout_logger,
)


def get_logger_for_stdout(output_format):
    """Gets the stdout logger for the given format.
        Args:
            output_format: CEF, JSON, or RAW_JSON. Each type results in a different logger instance.
    """
    formatter = _get_formatter(output_format)
    return get_stdout_logger(output_format.lower(), formatter)


def get_logger_for_file(filename, output_format):
    """Gets the logger that logs to a file for the given format.

        Args:
            filename: The name of the file to write logs to.
            output_format: CEF, JSON, or RAW_JSON. Each type results in a different logger instance.
    """
    logger = logging.getLogger(u"code42_file_{0}".format(output_format.lower()))
    if logger_has_handlers(logger):
        return logger

    with logger_deps_lock:
        if not logger_has_handlers(logger):
            handler = logging.FileHandler(filename, delay=True, encoding="utf-8")
            return _init_logger(logger, handler, output_format)
    return logger


def get_logger_for_server(hostname, protocol, output_format):
    """Gets the logger that sends logs to a server for the given format.

        Args:
            hostname: The hostname of the server. It may include the port.
            protocol: The transfer protocol for sending logs.
            output_format: CEF, JSON, or RAW_JSON. Each type results in a different logger instance.
    """
    logger = logging.getLogger(u"code42_syslog_{0}".format(output_format.lower()))
    if logger_has_handlers(logger):
        return logger

    with logger_deps_lock:
        if not logger_has_handlers(logger):
            url_parts = get_url_parts(hostname)
            port = url_parts[1] or 514
            try:
                handler = NoPrioritySysLogHandlerWrapper(
                    url_parts[0], port=port, protocol=protocol
                ).handler
            except:
                logger = get_main_cli_logger()
                logger.print_and_log_error(u"Unable to connect to {0}.".format(hostname))
                exit(1)
            return _init_logger(logger, handler, output_format)
    return logger


def _init_logger(logger, handler, output_format):
    formatter = _get_formatter(output_format)
    logger.setLevel(logging.INFO)
    return add_handler_to_logger(logger, handler, formatter)


def _get_formatter(output_format):
    if output_format == OutputFormat.JSON:
        return FileEventDictToJSONFormatter()
    elif output_format == OutputFormat.CEF:
        return FileEventDictToCEFFormatter()
    else:
        return FileEventDictToRawJSONFormatter()
